@@ -1,22 +1,21 @@
 # -*- coding: utf-8 -*-
-from operator import attrgetter
-import sys
-
 import django
 from django.conf import settings
-from django.db import connections
 from django.db import router
 from django.db import transaction
+from django.db import models
 from django.db.models import signals
-from django.db.models.fields.related import add_lazy_relation, create_many_related_manager, create_many_to_many_intermediary_model
-from django.db.models.fields.related import ManyToManyField, ReverseManyRelatedObjectsDescriptor
+from django.db.models.fields.related import add_lazy_relation
+from django.db.models.fields.related import ManyToManyField
 from django.db.models.fields.related import RECURSIVE_RELATIONSHIP_CONSTANT
 from django.utils import six
-from django.utils.functional import curry
+from django.utils.functional import cached_property, curry
 
+from .compat import create_forward_many_to_many_manager
+from .compat import get_foreignkey_field_kwargs
+from .compat import get_model_name
 from .forms import SortedMultipleChoiceField
 
-
 SORT_VALUE_FIELD_NAME = 'sort_value'
 
 
@@ -36,94 +35,9 @@ else:
             pass
 
 
-def create_sorted_many_to_many_intermediate_model(field, klass):
-    from django.db import models
-    managed = True
-    if isinstance(field.rel.to, six.string_types) and field.rel.to != RECURSIVE_RELATIONSHIP_CONSTANT:
-        to_model = field.rel.to
-        to = to_model.split('.')[-1]
-
-        def set_managed(field, model, cls):
-            field.rel.through._meta.managed = model._meta.managed or cls._meta.managed
-        add_lazy_relation(klass, field, to_model, set_managed)
-    elif isinstance(field.rel.to, six.string_types):
-        to = klass._meta.object_name
-        to_model = klass
-        managed = klass._meta.managed
-    else:
-        to = field.rel.to._meta.object_name
-        to_model = field.rel.to
-        managed = klass._meta.managed or to_model._meta.managed
-    name = '%s_%s' % (klass._meta.object_name, field.name)
-    if field.rel.to == RECURSIVE_RELATIONSHIP_CONSTANT or to == klass._meta.object_name:
-        from_ = 'from_%s' % to.lower()
-        to = 'to_%s' % to.lower()
-    else:
-        # Django 1.5 support.
-        if not hasattr(klass._meta, 'model_name'):
-            from_ = klass._meta.object_name.lower()
-        else:
-            from_ = klass._meta.model_name
-        to = to.lower()
-    options = {
-        'db_table': field._get_m2m_db_table(klass._meta),
-        'managed': managed,
-        'auto_created': klass,
-        'app_label': klass._meta.app_label,
-        'db_tablespace': klass._meta.db_tablespace,
-        'unique_together': (from_, to),
-        'ordering': (field.sort_value_field_name,),
-        'verbose_name': '%(from)s-%(to)s relationship' % {'from': from_, 'to': to},
-        'verbose_name_plural': '%(from)s-%(to)s relationships' % {'from': from_, 'to': to},
-    }
-    # Django 1.6 support.
-    if hasattr(field.model._meta, 'apps'):
-        options.update({
-            'apps': field.model._meta.apps,
-        })
-
-    meta = type(str('Meta'), (object,), options)
-    # Construct and return the new class.
-    def default_sort_value(name):
-        model = models.get_model(klass._meta.app_label, name)
-        # Django 1.5 support.
-        if django.VERSION < (1, 6):
-            return model._default_manager.count()
-        else:
-            from django.db.utils import ProgrammingError, OperationalError
-            try:
-                # We need to catch if the model is not yet migrated in the
-                # database. The default function is still called in this case while
-                # running the migration. So we mock the return value of 0.
-                with transaction.atomic():
-                    return model._default_manager.count()
-            except (ProgrammingError, OperationalError):
-                return 0
-
-    default_sort_value = curry(default_sort_value, name)
-
-    # Django 1.5 support.
-    if django.VERSION < (1, 6):
-        foreignkey_field_kwargs = {}
-    else:
-        foreignkey_field_kwargs = dict(
-            db_tablespace=field.db_tablespace,
-            db_constraint=field.rel.db_constraint)
-
-    return type(str(name), (models.Model,), {
-        'Meta': meta,
-        '__module__': klass.__module__,
-        from_: models.ForeignKey(klass, related_name='%s+' % name, **foreignkey_field_kwargs),
-        to: models.ForeignKey(to_model, related_name='%s+' % name, **foreignkey_field_kwargs),
-        field.sort_value_field_name: models.IntegerField(default=default_sort_value),
-        '_sort_field_name': field.sort_value_field_name,
-        '_from_field_name': from_,
-        '_to_field_name': to,
-    })
-
-
-def create_sorted_many_related_manager(superclass, rel):
-    RelatedManager = create_many_related_manager(superclass, rel)
+def create_sorted_many_related_manager(superclass, rel, *args, **kwargs):
+    RelatedManager = create_forward_many_to_many_manager(
+        superclass, rel, *args, **kwargs)
 
     class SortedRelatedManager(RelatedManager):
         def get_queryset(self):
@@ -173,13 +87,19 @@ def create_sorted_many_related_manager(superclass, rel):
 
         get_prefetch_query_set = get_prefetch_queryset
 
+        def set(self, objs, **kwargs):
+            # Choosing to clear first will ensure the order is maintained.
+            kwargs['clear'] = True
+            super(SortedRelatedManager, self).set(objs, **kwargs)
+        set.alters_data = True
+
         def _add_items(self, source_field_name, target_field_name, *objs):
             # source_field_name: the PK fieldname in join table for the source object
             # target_field_name: the PK fieldname in join table for the target object
             # *objs - objects to add. Either object instances, or primary keys of object instances.
 
             # If there aren't any objects, there is nothing to do.
-            from django.db.models import Model
+            from django.db.models import Max, Model
             if objs:
                 # Django uses a set here, we need to use a list to keep the
                 # correct ordering.
@@ -208,7 +128,8 @@ def create_sorted_many_related_manager(superclass, rel):
                         new_ids.append(obj)
 
                 db = router.db_for_write(self.through, instance=self.instance)
-                vals = (self.through._default_manager.using(db)
+                manager = self.through._default_manager.using(db)
+                vals = (manager
                         .values_list(target_field_name, flat=True)
                         .filter(**{
                             source_field_name: self._fk_val,
@@ -224,39 +145,67 @@ def create_sorted_many_related_manager(superclass, rel):
                 new_ids = _new_ids
                 new_ids_set = set(new_ids)
 
-                with atomic(using=db, savepoint=False):
-                    if self.reverse or source_field_name == self.source_field_name:
-                        # Don't send the signal when we are inserting the
-                        # duplicate data row for symmetrical reverse entries.
-                        signals.m2m_changed.send(sender=rel.through, action='pre_add',
-                            instance=self.instance, reverse=self.reverse,
-                            model=self.model, pk_set=new_ids_set, using=db)
-                    # Add the ones that aren't there already
+                if self.reverse or source_field_name == self.source_field_name:
+                    # Don't send the signal when we are inserting the
+                    # duplicate data row for symmetrical reverse entries.
+                    signals.m2m_changed.send(sender=rel.through, action='pre_add',
+                        instance=self.instance, reverse=self.reverse,
+                        model=self.model, pk_set=new_ids_set, using=db)
+
+                # Add the ones that aren't there already
+                with atomic(using=db):
+                    fk_val = self._fk_val
+                    source_queryset = manager.filter(**{'%s_id' % source_field_name: fk_val})
                     sort_field_name = self.through._sort_field_name
-                    sort_field = self.through._meta.get_field_by_name(sort_field_name)[0]
-                    for obj_id in new_ids:
-                        self.through._default_manager.using(db).create(**{
-                            '%s_id' % source_field_name: self._fk_val,  # Django 1.5 compatibility
-                            '%s_id' % target_field_name: obj_id,
-                            sort_field_name: sort_field.get_default(),
+                    sort_value_max = source_queryset.aggregate(max=Max(sort_field_name))['max'] or 0
+
+                    manager.bulk_create([
+                        self.through(**{
+                            '%s_id' % source_field_name: fk_val,
+                            '%s_id' % target_field_name: pk,
+                            sort_field_name: sort_value_max + i + 1,
                         })
-                    if self.reverse or source_field_name == self.source_field_name:
-                        # Don't send the signal when we are inserting the
-                        # duplicate data row for symmetrical reverse entries.
-                        signals.m2m_changed.send(sender=rel.through, action='post_add',
-                            instance=self.instance, reverse=self.reverse,
-                            model=self.model, pk_set=new_ids_set, using=db)
+                        for i, pk in enumerate(new_ids)
+                    ])
+
+                if self.reverse or source_field_name == self.source_field_name:
+                    # Don't send the signal when we are inserting the
+                    # duplicate data row for symmetrical reverse entries.
+                    signals.m2m_changed.send(sender=rel.through, action='post_add',
+                        instance=self.instance, reverse=self.reverse,
+                        model=self.model, pk_set=new_ids_set, using=db)
 
     return SortedRelatedManager
 
 
-class ReverseSortedManyRelatedObjectsDescriptor(ReverseManyRelatedObjectsDescriptor):
-    @property
-    def related_manager_cls(self):
-        return create_sorted_many_related_manager(
-            self.field.rel.to._default_manager.__class__,
-            self.field.rel
-        )
+try:
+    from django.db.models.fields.related_descriptors import ManyToManyDescriptor
+
+    class SortedManyToManyDescriptor(ManyToManyDescriptor):
+        def __init__(self, field):
+            super(SortedManyToManyDescriptor, self).__init__(field.remote_field)
+
+        @cached_property
+        def related_manager_cls(self):
+            model = self.rel.model
+            return create_sorted_many_related_manager(
+                model._default_manager.__class__,
+                self.rel,
+                # This is the new `reverse` argument (which ironically should
+                # be False)
+                reverse=False,
+            )
+except ImportError:
+    # Django 1.8 support
+    from django.db.models.fields.related import ReverseManyRelatedObjectsDescriptor
+
+    class SortedManyToManyDescriptor(ReverseManyRelatedObjectsDescriptor):
+        @cached_property
+        def related_manager_cls(self):
+            return create_sorted_many_related_manager(
+                self.field.rel.to._default_manager.__class__,
+                self.field.rel
+            )
 
 
 class SortedManyToManyField(ManyToManyField):
@@ -273,13 +222,24 @@ class SortedManyToManyField(ManyToManyField):
         self.sort_value_field_name = kwargs.pop(
             'sort_value_field_name',
             SORT_VALUE_FIELD_NAME)
+
         super(SortedManyToManyField, self).__init__(to, **kwargs)
         if self.sorted:
             self.help_text = kwargs.get('help_text', None)
 
-    def contribute_to_class(self, cls, name):
+    def deconstruct(self):
+        # We have to persist custom added options in the ``kwargs``
+        # dictionary. For readability only non-default values are stored.
+        name, path, args, kwargs = super(SortedManyToManyField, self).deconstruct()
+        if self.sort_value_field_name is not SORT_VALUE_FIELD_NAME:
+            kwargs['sort_value_field_name'] = self.sort_value_field_name
+        if self.sorted is not True:
+            kwargs['sorted'] = self.sorted
+        return name, path, args, kwargs
+
+    def contribute_to_class(self, cls, name, **kwargs):
         if not self.sorted:
-            return super(SortedManyToManyField, self).contribute_to_class(cls, name)
+            return super(SortedManyToManyField, self).contribute_to_class(cls, name, **kwargs)
 
         # To support multiple relations to self, it's useful to have a non-None
         # related name on symmetrical relations for internal reasons. The
@@ -290,16 +250,16 @@ class SortedManyToManyField(ManyToManyField):
         if self.rel.symmetrical and (self.rel.to == "self" or self.rel.to == cls._meta.object_name):
             self.rel.related_name = "%s_rel_+" % name
 
-        super(ManyToManyField, self).contribute_to_class(cls, name)
+        super(ManyToManyField, self).contribute_to_class(cls, name, **kwargs)
 
         # The intermediate m2m model is not auto created if:
         #  1) There is a manually specified intermediate, or
         #  2) The class owning the m2m field is abstract.
         if not self.rel.through and not cls._meta.abstract:
-            self.rel.through = create_sorted_many_to_many_intermediate_model(self, cls)
+            self.rel.through = self.create_intermediate_model(cls)
 
         # Add the descriptor for the m2m relation
-        setattr(cls, self.name, ReverseSortedManyRelatedObjectsDescriptor(self))
+        setattr(cls, self.name, SortedManyToManyDescriptor(self))
 
         # Set up the accessor for the m2m table name for the relation
         self.m2m_db_table = curry(self._get_m2m_db_table, cls._meta)
@@ -318,6 +278,9 @@ class SortedManyToManyField(ManyToManyField):
                 target = self.rel.to._meta.db_table
             cls._meta.duplicate_targets[self.column] = (target, "m2m")
 
+    def get_internal_type(self):
+        return 'ManyToManyField'
+
     def formfield(self, **kwargs):
         defaults = {}
         if self.sorted:
@@ -325,6 +288,117 @@ class SortedManyToManyField(ManyToManyField):
         defaults.update(kwargs)
         return super(SortedManyToManyField, self).formfield(**defaults)
 
+    ###############################
+    # Intermediate Model Creation #
+    ###############################
+
+    def get_intermediate_model_name(self, klass):
+        return '%s_%s' % (klass._meta.object_name, self.name)
+
+    def get_intermediate_model_meta_class(self, klass, from_field_name,
+                                          to_field_name,
+                                          sort_value_field_name):
+        managed = True
+        to_model = self.rel.to
+        if isinstance(self.rel.to, six.string_types):
+            if self.rel.to != RECURSIVE_RELATIONSHIP_CONSTANT:
+                def set_managed(field, model, cls):
+                    field.rel.through._meta.managed = model._meta.managed or cls._meta.managed
+                add_lazy_relation(klass, self, to_model, set_managed)
+            else:
+                managed = klass._meta.managed
+        else:
+            managed = klass._meta.managed or to_model._meta.managed
+
+        options = {
+            'db_table': self._get_m2m_db_table(klass._meta),
+            'managed': managed,
+            'auto_created': klass,
+            'app_label': klass._meta.app_label,
+            'db_tablespace': klass._meta.db_tablespace,
+            'unique_together': ((from_field_name, to_field_name),),
+            'ordering': (sort_value_field_name,),
+            'verbose_name': '%(from)s-%(to)s relationship' % {'from': from_field_name, 'to': to_field_name},
+            'verbose_name_plural': '%(from)s-%(to)s relationships' % {'from': from_field_name, 'to': to_field_name},
+        }
+        # Django 1.6 support.
+        if hasattr(self.model._meta, 'apps'):
+            options.update({
+                'apps': self.model._meta.apps,
+            })
+        return type(str('Meta'), (object,), options)
+
+    def get_rel_to_model_and_object_name(self, klass):
+        if isinstance(self.rel.to, six.string_types):
+            if self.rel.to != RECURSIVE_RELATIONSHIP_CONSTANT:
+                to_model = self.rel.to
+                to_object_name = to_model.split('.')[-1]
+            else:
+                to_model = klass
+                to_object_name = to_model._meta.object_name
+        else:
+            to_model = self.rel.to
+            to_object_name = to_model._meta.object_name
+        return to_model, to_object_name
+
+    def get_intermediate_model_sort_value_field(self, klass):
+        field_name = self.sort_value_field_name
+        field = models.IntegerField(default=0)
+        return field_name, field
+
+    def get_intermediate_model_from_field(self, klass):
+        name = self.get_intermediate_model_name(klass)
+
+        to_model, to_object_name = self.get_rel_to_model_and_object_name(klass)
+
+        if self.rel.to == RECURSIVE_RELATIONSHIP_CONSTANT or to_object_name == klass._meta.object_name:
+            field_name = 'from_%s' % to_object_name.lower()
+        else:
+            field_name = get_model_name(klass)
+
+        field = models.ForeignKey(klass, related_name='%s+' % name,
+                                  **get_foreignkey_field_kwargs(self))
+        return field_name, field
+
+    def get_intermediate_model_to_field(self, klass):
+        name = self.get_intermediate_model_name(klass)
+
+        to_model, to_object_name = self.get_rel_to_model_and_object_name(klass)
+
+        if self.rel.to == RECURSIVE_RELATIONSHIP_CONSTANT or to_object_name == klass._meta.object_name:
+            field_name = 'to_%s' % to_object_name.lower()
+        else:
+            field_name = to_object_name.lower()
+
+        field = models.ForeignKey(to_model, related_name='%s+' % name,
+                                  **get_foreignkey_field_kwargs(self))
+        return field_name, field
+
+    def create_intermediate_model_from_attrs(self, klass, attrs):
+        name = self.get_intermediate_model_name(klass)
+        return type(str(name), (models.Model,), attrs)
+
+    def create_intermediate_model(self, klass):
+        # Construct and return the new class.
+        from_field_name, from_field = self.get_intermediate_model_from_field(klass)
+        to_field_name, to_field = self.get_intermediate_model_to_field(klass)
+        sort_value_field_name, sort_value_field = self.get_intermediate_model_sort_value_field(klass)
+
+        meta = self.get_intermediate_model_meta_class(
+            klass, from_field_name, to_field_name, sort_value_field_name)
+
+        attrs = {
+            'Meta': meta,
+            '__module__': klass.__module__,
+            from_field_name: from_field,
+            to_field_name: to_field,
+            sort_value_field_name: sort_value_field,
+            '_sort_field_name': sort_value_field_name,
+            '_from_field_name': from_field_name,
+            '_to_field_name': to_field_name,
+        }
+        return self.create_intermediate_model_from_attrs(klass, attrs)
+
 
 # Add introspection rules for South database migrations
 # See http://south.aeracode.org/docs/customfields.html
@@ -333,6 +407,7 @@ try:
 except ImportError:
     south = None
 
+
 if south is not None and 'south' in settings.INSTALLED_APPS:
     from south.modelsinspector import add_introspection_rules
     add_introspection_rules(
