@@ -4,29 +4,41 @@ if (jQuery === undefined) {
 
 (function ($) {
     $(function () {
-        $('.sortedm2m').parents('ul').each(function () {
-            $(this).addClass('sortedm2m');
-            var checkboxes = $(this).find('input[type=checkbox]');
+        $('.sortedm2m-container').find('ul').addClass('hide');
+        function prepareUl(ul) {
+            ul.addClass('sortedm2m');
+            var checkboxes = ul.find('input[type=checkbox]');
             var id = checkboxes.first().attr('id').match(/^(.*)_\d+$/)[1];
             var name = checkboxes.first().attr('name');
             checkboxes.removeAttr('name');
-            $(this).before('<input type="hidden" id="' + id + '" name="' + name + '" />');
-            var that = this;
+            ul.before('<input type="hidden" id="' + id + '" name="' + name + '" />');
             var recalculate_value = function () {
                 var values = [];
-                $(that).find(':checked').each(function () {
+                ul.find(':checked').each(function () {
                     values.push($(this).val());
                 });
                 $('#' + id).val(values.join(','));
-            }
+            };
             recalculate_value();
-            checkboxes.change(recalculate_value);
-            $(this).sortable({
+            ul.on('change','input[type=checkbox]',recalculate_value);
+            ul.sortable({
                 axis: 'y',
                 //containment: 'parent',
                 update: recalculate_value
             });
-        });
+            return ul;
+        }
+
+        function iterateUl() {
+            $('ul:has(.sortedm2m)').each(function () {
+                prepareUl( $(this) );
+                $(this).removeClass('hide');
+            });
+        }
+
+        $(".add-row a").click(iterateUl);
+
+        iterateUl();
 
         $('.sortedm2m-container .selector-filter input').each(function () {
             $(this).bind('input', function() {
@@ -40,15 +52,21 @@ if (jQuery === undefined) {
                             $(this).css('display', 'none');
                         } else {
                             $(this).css('display', 'inherit');
-                        };
+                        }
                     });
                 });
             });
         });
 
+        var dismissPopupFnName = 'dismissAddAnotherPopup';
+        // django 1.8+
+        if (window.dismissAddRelatedObjectPopup) {
+            dismissPopupFnName = 'dismissAddRelatedObjectPopup';
+        }
+
         if (window.showAddAnotherPopup) {
-            var django_dismissAddAnotherPopup = window.dismissAddAnotherPopup;
-            window.dismissAddAnotherPopup = function (win, newId, newRepr) {
+            var django_dismissAddAnotherPopup = window[dismissPopupFnName];
+            window[dismissPopupFnName] = function (win, newId, newRepr) {
                 // newId and newRepr are expected to have previously been escaped by
                 // django.utils.html.escape.
                 newId = html_unescape(newId);
